/*  CAO Compiler
    Copyright (C) 2014 Cryptography and Information Security Group, HASLab - INESC TEC and Universidade do Minho

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef CAO_MATRIX_H
#define CAO_MATRIX_H

#include "CAO_globals.h"
#include "CAO_globalOp.h"

#ifdef __cplusplus

#include <NTL/ZZ.h>
#include <iostream>

NTL_CLIENT

typedef struct CAO_matrix_s {
	int rows;
	int cols;
	char type;
	CAO_REF *value;
} CAO_matrix_s;

extern "C" {
#endif

	CAO_RES CAO_matrix_decl(CAO_matrix *, CAO_rint, CAO_rint, const char *, void **);
	CAO_RES _CAO_matrix_decl(CAO_matrix *, CAO_rint, CAO_rint, const char *, void **, int *);

	CAO_RES CAO_matrix_dispose(CAO_matrix);
	CAO_RES CAO_matrix_const_init(CAO_matrix, void *);

	CAO_RES CAO_matrix_init(CAO_matrix, void **);
	CAO_RES _CAO_matrix_init(CAO_matrix, void **, int *);

	CAO_RES CAO_matrix_assign(CAO_matrix, CAO_matrix);
	CAO_RES CAO_matrix_clone(CAO_matrix *, CAO_matrix);

	CAO_REF CAO_matrix_ref(CAO_matrix, CAO_rint, CAO_rint);
	CAO_REF _CAO_matrix_ref(CAO_matrix, CAO_rint, CAO_rint, char *);

	CAO_RES CAO_matrix_dump(CAO_matrix);

	#define CAO_matrix_equal(a,b,c) a = _CAO_matrix_equal(b,c)
	CAO_bool _CAO_matrix_equal(CAO_matrix, CAO_matrix);

	CAO_RES CAO_matrix_select(CAO_REF, CAO_matrix, CAO_rint, CAO_rint);
	CAO_RES CAO_matrix_range_select(CAO_matrix, CAO_matrix, CAO_rint, CAO_rint, CAO_rint, CAO_rint);
	CAO_RES CAO_matrix_range_set(CAO_matrix, CAO_matrix, CAO_rint, CAO_rint, CAO_rint, CAO_rint);
	CAO_RES CAO_matrix_row_range_select(CAO_matrix r, CAO_matrix m, CAO_rint c, CAO_rint ri, CAO_rint rj);
	CAO_RES CAO_matrix_col_range_select (CAO_matrix r, CAO_matrix m, CAO_rint row, CAO_rint ci, CAO_rint cj);
	CAO_RES CAO_matrix_row_range_set (CAO_matrix r, CAO_matrix m, CAO_rint c, CAO_rint ri, CAO_rint rj);
	CAO_RES CAO_matrix_col_range_set (CAO_matrix r, CAO_matrix m, CAO_rint row, CAO_rint ci, CAO_rint cj);

	CAO_RES CAO_matrix_addTo (CAO_matrix r, CAO_matrix m8);
	CAO_RES CAO_matrix_add (CAO_matrix r, CAO_matrix a, CAO_matrix b);
	CAO_RES CAO_matrix_subTo (CAO_matrix r, CAO_matrix m);
	CAO_RES CAO_matrix_sub (CAO_matrix r, CAO_matrix a, CAO_matrix b);
	CAO_RES CAO_matrix_sym (CAO_matrix r, CAO_matrix m);
	CAO_RES CAO_matrix_mul (CAO_matrix r, CAO_matrix a, CAO_matrix b);
	CAO_RES CAO_matrix_assign_zero (CAO_matrix);
	CAO_RES CAO_matrix_assign_one (CAO_matrix);
	CAO_RES CAO_matrix_pow (CAO_matrix r, CAO_matrix m, CAO_int n);

	CAO_RES CAO_matrix_cast_matrix (CAO_matrix, CAO_matrix);
	
#ifdef __cplusplus
}
#endif

#endif
