/*  CAO Compiler
    Copyright (C) 2014 Cryptography and Information Security Group, HASLab - INESC TEC and Universidade do Minho

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef CAO_UBITS_H
#define CAO_UBITS_H

#include "CAO_globals.h"
#include "CAO_bool.h"
#include "CAO_int.h"
#include "CAO_mod.h"

#ifdef __cplusplus

#include <NTL/ZZ.h>
#include <iostream>

NTL_CLIENT

typedef struct CAO_ubits_s {
	int size;
	ZZ *value;
} CAO_ubits_s;

extern "C" {
#endif

	CAO_RES CAO_ubits_decl(CAO_ubits*, CAO_rint);
	CAO_RES CAO_ubits_init(CAO_ubits, const char *);
	CAO_RES CAO_ubits_assign(CAO_ubits ,CAO_ubits);
	CAO_RES CAO_ubits_clone(CAO_ubits *, CAO_ubits);
	CAO_RES CAO_ubits_dispose(CAO_ubits);

	#define CAO_ubits_equal(a,b,c) a = _CAO_ubits_equal(b,c)
	CAO_bool _CAO_ubits_equal(CAO_ubits, CAO_ubits);
	#define CAO_ubits_nequal(a,b,c) a = _CAO_ubits_nequal(b,c)
	CAO_bool _CAO_ubits_nequal(CAO_ubits, CAO_ubits);

	CAO_RES CAO_ubits_not(CAO_ubits, CAO_ubits);
	CAO_RES CAO_ubits_or(CAO_ubits, CAO_ubits, CAO_ubits);
	CAO_RES CAO_ubits_xor(CAO_ubits, CAO_ubits, CAO_ubits);
	CAO_RES CAO_ubits_and(CAO_ubits, CAO_ubits, CAO_ubits);
	CAO_RES CAO_ubits_shift_up(CAO_ubits, CAO_ubits, CAO_rint);
	CAO_RES CAO_ubits_shift_down(CAO_ubits, CAO_ubits, CAO_rint);
	CAO_RES CAO_ubits_rot_up(CAO_ubits, CAO_ubits, CAO_rint);
	CAO_RES CAO_ubits_rot_down(CAO_ubits, CAO_ubits, CAO_rint);
	CAO_RES CAO_ubits_range_select(CAO_ubits, CAO_ubits, CAO_rint, CAO_rint);
	CAO_RES CAO_ubits_select(CAO_ubits, CAO_ubits, CAO_rint);
	CAO_RES CAO_ubits_range_set(CAO_ubits, CAO_ubits, CAO_rint, CAO_rint);
	CAO_RES CAO_ubits_set(CAO_ubits, CAO_ubits, CAO_rint);
	CAO_RES CAO_ubits_concat(CAO_ubits, CAO_ubits, CAO_ubits);
	CAO_RES CAO_ubits_dump(CAO_ubits);

	CAO_RES CAO_ubits_cast_int(CAO_int, CAO_ubits);
	CAO_RES CAO_int_cast_ubits(CAO_ubits, CAO_int);
	CAO_RES CAO_ubits_cast_mod(CAO_int, CAO_ubits);
	CAO_RES CAO_mod_cast_ubits(CAO_ubits, CAO_int);
	CAO_RES CAO_ubits_cast_ubits(CAO_ubits, CAO_ubits);

#ifdef __cplusplus
}
#endif

#endif
